/*
 * Styles for widget display
 */

.widget-area .prompt, .output_stderr {
    display: none;
}

.widget-float-right {
    float: right;
    position: relative;
    top: -3px;
    left: 7px;
}

.gp-widget {
    width: 100%;
    max-width: 100%;
    border-color: rgba(10, 45, 105, 0.80);
}

.gp-widget .panel-heading {
    background-color: rgba(10, 45, 105, 0.80);
    color: white;
}

.gp-widget label {
    font-weight: bold;
}

.gp-widget-logo {
    height: 20px;
    float: left;
    padding-right: 10px;
    position: relative;
    left: -5px;
    top: -2px;
}

.widget-username-label {
    font-style: italic;
}

.widget-server-label {
    font-size: 0.8em;
    font-style: italic;
    padding-right: 10px;
}

.gp-widget .btn-primary {
    border-color: rgba(10, 45, 105, 0.80);
    background-color: rgba(10, 45, 105, 0.80);
}

.gp-widget .btn-primary:hover {
    border-color: rgba(10, 45, 105, 1);
    background-color: rgba(10, 45, 105, 1);
}

.widget-view {
    position: relative;
}

.widget-subarea {
    max-width: 100%;
}

.gp-widget-loading, .gp-widget-logged-in {
    position: absolute;
    text-align: center;
    top: 0;
    bottom: 0;
    right: 0;
    left: 0;
    background: rgba(255,255,255,0.75);
}

.gp-widget-loading img {
    width: 100px;
    position: relative;
    top: 75px;
}

.gp-widget-logged-in div {
    position: relative;
    top: 75px;
    display: inline-block;
    line-height: 40px;
    font-size: 1.2em;
}

.widget-code {
    -moz-box-align: stretch;
    -moz-box-orient: horizontal;
    align-items: stretch;
    display: flex;
    flex-direction: row;
    page-break-inside: avoid;
}

.widget-code .prompt {
    display: inline-block;
}

.widget-code pre {
    border: none;
    border-radius: 0;
    padding: 0;
    margin: 0;
}

/*
 * Job Widget Styles
 */

.gp-widget-job-status {
    padding-bottom: 3px;
}

.gp-widget-job-submitted {
    font-style: italic;
    font-size:0.9em;
}

.gp-widget-job-buttons {
    text-align: right;
}

.gp-widget-job-share-table {
    width: 50%;
}

.gp-widget-job-share-table input[type=radio] {
    margin-left: 20px;
    margin-right: 5px;
}

.gp-widget-job-child {
    border-color: #BBBBBB;
}

.gp-widget-job-child .gp-widget-job-header {
    background-color: #BBBBBB;
    border-color: #BBBBBB;
}

/*
 * File Input Styles
 */

.file-widget {
    border: 2px dashed darkgray;
    display: inline-block;
    min-width: 450px;
    width: 100%;
    padding: 9px 10px 10px;
    white-space: normal;
    word-break: break-all;
}

.file-widget-upload-file {
    margin-right: 7px;
}

.file-widget-input-file {
    display: none;
}

.file-widget-input-file {
    position: absolute;
    z-index: -1000;
}

.file-widget-drop {
    color: darkgray;
    font-size: 2em;
    letter-spacing: 1px;
    margin-left: 9px;
    padding: 5px 10px;
    position: relative;
    top: 10px;
}

.file-widget-size {
    font-size: 0.9em;
    font-style: italic;
    margin-left: 5px;
    margin-top: 5px;
}

.file-widget-value {
    background-color: #f7f7f7;
    margin-top: 10px;
    overflow: hidden;
    padding: 4px;
    white-space: normal;
    line-height: 2.2em;
    vertical-align: middle;
}

.file-widget-value-erase {
    float: right;
}

.file-widget-path-label {
    font-weight: bold;
    text-align: left !important;
}

.file-widget-path-buttons {
    margin-top: 5px;
    text-align: right;
}

/*
 * Choice Input Styles
 */

.choice-widget-select {
    position: relative;
    left: -10px;
}

/*
 * Run Task Styles
 */

.gp-widget-task-version {
    font-size: 0.8em;
    font-style: italic;
    padding-right: 10px;
}

.gp-widget-task-subheader {
    border-bottom: 1px solid rgba(10, 45, 105, 0.8);
}

.gp-widget-task-run {
    overflow: auto;
    margin-left: 20px;
    font-size: 0.9em;
    padding-top: 10px;
}

.gp-widget-task-run-button {
    float: right;
    position: relative;
    top: -10px;
}

.gp-widget-task-form {
    padding-top: 10px;
    padding-bottom: 10px;
}

.gp-widget-task-param {
    margin: 10px;
    padding: 10px;
}

.gp-widget-task-param-name {
    display: inline-block;
    /*width: 150px;*/
    font-weight: bold;
    font-size: 1em;
    word-break: break-all;
    vertical-align: top;
    padding-top: 10px;
    padding-right: 10px;
}

.gp-widget-task-param-wrapper {
    vertical-align: top;
    padding-left: 10px;
}

.gp-widget-task-param-desc {
    font-size: 1em;
    margin-top: 10px;
    margin-bottom: 10px;
}

.gp-widget-task-param-missing {
    border: 2px solid red;
}

.gp-widget-task-footer {
    border-top: 1px solid rgba(10, 45, 105, 0.8);
    padding-top: 10px;
}

.gp-widget-task-eula-license {
    width: 100%;
    height: 300px;
    overflow: auto;
    text-align: left;
    font-size: 0.8em;
    font-weight: normal;
    line-height: 1.2em;
}

.gp-widget-task-eula {
    padding: 20px;
}

.gp-widget-task-eula div {
    width: 100%;
    font-weight: bold;
    line-height: 3em;
}